import aiohttp
import logging
from http import HTTPStatus

from ..utils import json, exceptions
from ..utils.helper import Helper, HelperMode, Item

BASE_URL = 'https://dialogs.yandex.net/api/v1/'
API_URL = BASE_URL + 'skills/{skill_id}/{method}/'

log = logging.getLogger(__name__)


async def _check_result(response):
    body = await response.text()
    if response.content_type != 'application/json':
        log.error('Invalid response with content type %r: %r',
                  response.content_type, body)
        exceptions.DialogsAPIError.detect(body)

    try:
        result_json = await response.json(loads=json.loads)
    except ValueError:
        result_json = {}

    log.debug('Request result is %r', result_json)

    if HTTPStatus.OK <= response.status <= HTTPStatus.IM_USED:
        return result_json
    if result_json and 'message' in result_json:
        description = result_json['message']
    else:
        description = body

    log.warning('Response status %r with description %r',
                response.status, description)
    exceptions.DialogsAPIError.detect(description)


async def request(session, oauth_token, skill_id=None, method=None, json=None,
                  file=None, request_method='POST', custom_url=None, **kwargs):
    """
    Make a request to API

    :param session: HTTP Client session
    :type session: :obj:`aiohttp.ClientSession`
    :param oauth_token: oauth_token
    :type oauth_token: :obj:`str`
    :param skill_id: skill_id. Optional. Not used if custom_url is provided
    :type skill_id: :obj:`str`
    :param method: API method. Optional. Not used if custom_url is provided
    :type method: :obj:`str`
    :param json: request payload
    :type json: :obj: `dict`
    :param file: file
    :type file: :obj: `io.BytesIO`
    :param request_method: API request method
    :type request_method: :obj:`str`
    :param custom_url: Yandex has very developer UNfriendly API, so some endpoints cannot be achieved by standatd template.
    :type custom_url: :obj:`str`
    :return: result
    :rtype: ::obj:`dict`
    """
    log.debug("Making a `%s` request to %r with json `%r` or file `%r`",
              request_method, method, json, file)
    if custom_url is None:
        url = Methods.api_url(skill_id, method)
    else:
        url = custom_url
    headers = {'Authorization': oauth_token}
    data = None
    if file:
        data = aiohttp.FormData()
        data.add_field('file', file)
    try:
        async with session.request(request_method, url, json=json, data=data, headers=headers, **kwargs) as response:
            return await _check_result(response)
    except aiohttp.ClientError as e:
        raise exceptions.NetworkError(f"aiohttp client throws an error: {e.__class__.__name__}: {e}")


class Methods(Helper):

    mode = HelperMode.lowerCamelCase

    IMAGES = Item()  # images
    STATUS = Item()  # status

    @staticmethod
    def api_url(skill_id, method):
        """
        Generate API URL with skill_id and method

        :param skill_id:
        :param method:
        :return:
        """
        return API_URL.format(skill_id=skill_id, method=method)



import aiohttp
import asyncio
# import logging

from . import api
from .handler import Handler, SkipHandler
from .storage import DisabledStorage, MemoryStorage, DEFAULT_STATE
from .filters import generate_default_filters, ExceptionsFilter
from ..utils import json, exceptions

from ..types import UploadedImage, Quota
# log = logging.getLogger(__name__)


class Dispatcher:

    def __init__(self, loop=None, storage=None, *, skill_id=None, oauth_token=None):
        # TODO: inculde default handler for 'test' commands
        # TODO: create default handler for exceptions handler
        self.loop = loop or asyncio.get_event_loop()
        self.storage = storage or DisabledStorage()
        self.requests_handlers = Handler()
        self.errors_handlers = Handler()

        self.skill_id = skill_id
        self.oauth_token = oauth_token

        self.__session = None  # Lazy initialize session

    @property
    def session(self):
        if self.__session is None:
            self.__session = aiohttp.ClientSession(
                loop=self.loop, json_serialize=json.dumps
            )
        return self.__session

    async def close(self):
        """
        Close all client sessions

        If doing any requests outside of web app don't forget
        to close session manually by calling `await dp.close`
        """
        if self.__session and not self.__session.closed:
            await self.__session.close()

    async def shutdown(self, webapp):
        await self.close()

    async def process_request(self, request):
        try:
            return await self.requests_handlers.notify(request)
        except Exception as e:
            result = await self.errors_handlers.notify(request, e)
            if result:
                return result
            raise

    def register_request_handler(self, callback, *, commands=None, contains=None, starts_with=None, request_type=None,
                                 func=None, state=DEFAULT_STATE, regexp=None, custom_filters=None, **kwargs):
        """
        Register handler for AliceRequest

        .. code-block:: python3
            dp = Dispatcher()

            async def handle_all_requests(alice_request):
                return alice_request.response('Hello world!')

            dp.register_request_handler(handle_all_requests)


        :param callback: function to process request
        :param commands: list of commands
        :param contains: list of lines to search in commands `any([line in command for line in contains])`
        :param starts_with: list of lines to check if command starts with any of them
        :param request_type: Type of the request can be 'SimpleUtterance' or 'ButtonPressed'
        :param func: any callable object (for custom checks)
        :param state: For FSM
        :param regexp: REGEXP
        :param custom_filters: list of custom filters
        :param kwargs:
        """
        if custom_filters is None:
            custom_filters = []

        prepared_filers = generate_default_filters(
            self, *custom_filters, commands=commands, contains=contains,
            starts_with=starts_with, request_type=request_type,
            func=func, state=state, regexp=regexp, **kwargs
        )
        self.requests_handlers.register(callback, prepared_filers)

    def request_handler(self, *custom_filters, commands=None, contains=None, starts_with=None,
                        request_type=None, func=None, state=DEFAULT_STATE, regexp=None, **kwargs):
        """
        Decorator AliceRequest handler

        .. code-block:: python3
            dp = Dispatcher()

            @dp.request_handler()
            async def handle_all_requests(alice_request):
                return alice_request.response('Hello world!')


        :param callback: function to process request
        :param commands: list of commands
        :param contains: list of lines to search in commands `any([line in command for line in contains])`
        :param starts_with: list of lines to check if command starts with any of them
        :param request_type: Type of the request can be 'SimpleUtterance' or 'ButtonPressed'
        :param func: any callable object (for custom checks)
        :param state: For FSM
        :param regexp: REGEXP
        :param custom_filters: list of custom filters
        :param kwargs:
        :return: decorated function
        """
        def decorator(callback):
            self.register_request_handler(
                callback, commands=commands, contains=contains, starts_with=starts_with, request_type=request_type,
                func=func, state=state, regexp=regexp, custom_filters=custom_filters, **kwargs
            )
            return callback

        return decorator

    def register_errors_handler(self, callback, *, exception=None, func=None):
        """
        Register handler for errors

        :param callback:
        :param exception: you can make handler for specific errors type
        :param func: any callable object (for custom checks)
        """
        filters_list = []
        if func is not None:
            filters_list.append(func)
        if exception is not None:
            filters_list.append(ExceptionsFilter(exception))
        self.errors_handlers.register(callback, filters_list)

    def errors_handler(self, exception=None, func=None):
        """
        Decorator for errors handler

        :param func:
        :param exception: you can make handler for specific errors type
        :param run_task: run callback in task (no wait results)
        :return: decorated function
        """

        def decorator(callback):
            self.register_errors_handler(callback, exception=exception, func=func)
            return callback

        return decorator

    def _check_auth(self, skill_id, oauth_token):
        skill_id = skill_id or self.skill_id
        oauth_token = oauth_token or self.oauth_token
        if not (skill_id and oauth_token):
            raise exceptions.AuthRequired('Please provide both skill_id and oauth_token')
        return skill_id, oauth_token

    async def get_images(self, skill_id=None, oauth_token=None):
        '''
        Get uploaded images

        :param skill_id: Provide if was not set at the Dispatcher init
        :type skill_id: :obj:`str`
        :param oauth_token: Provide if was not set at the Dispatcher init
        :type oauth_token: :obj:`str`

        :return: list of UploadedImage instances
        '''
        skill_id, oauth_token = self._check_auth(skill_id, oauth_token)
        result = await api.request(
            self.session, oauth_token, skill_id,
            api.Methods.IMAGES, request_method='GET'
        )
        if 'images' not in result:
            raise exceptions.ApiChanged(f'Expected "images" in result, got {result}')
        return [UploadedImage(**dct) for dct in result['images']]

    async def upload_image(self, image_url_or_bytes, skill_id=None, oauth_token=None):
        '''
        Upload image by either url or bytes

        :param image_url_or_bytes: Image URL or bytes
        :type image_url_or_bytes: :obj:`str` or `io.BytesIO`
        :param skill_id: Provide if was not set at the Dispatcher init
        :type skill_id: :obj:`str`
        :param oauth_token: Provide if was not set at the Dispatcher init
        :type oauth_token: :obj:`str`

        :return: UploadedImage
        '''
        skill_id, oauth_token = self._check_auth(skill_id, oauth_token)
        json = None
        file = None
        if isinstance(image_url_or_bytes, str):
            json = {'url': image_url_or_bytes}
        else:
            file = image_url_or_bytes
        result = await api.request(
            self.session, oauth_token, skill_id,
            api.Methods.IMAGES, json, file
        )
        if 'image' not in result:
            raise exceptions.ApiChanged(f'Expected "image" in result, got {result}')
        return UploadedImage(**result['image'])

    async def get_images_quota(self, oauth_token=None):
        '''
        Get images storage quota

        :param oauth_token: Provide if was not set at the Dispatcher init
        :type oauth_token: :obj:`str`

        :return: Quota
        '''
        oauth_token = oauth_token or self.oauth_token
        if oauth_token is None:
            raise exceptions.AuthRequired('Please provide oauth_token')

        result = await api.request(
            self.session, oauth_token, request_method='GET',
            custom_url=api.BASE_URL + api.Methods.STATUS
        )
        if 'images' not in result or 'quota' not in result['images']:
            raise exceptions.ApiChanged(f'Expected "images" "quota" in result, got {result}')
        return Quota(**result['images']['quota'])

    async def delete_image(self, image_id, skill_id=None, oauth_token=None):
        '''
        Delete image by id

        :param image_id: Image id to be deleted
        :type image_id: :obj:`str`
        :param skill_id: Provide if was not set at the Dispatcher init
        :type skill_id: :obj:`str`
        :param oauth_token: Provide if was not set at the Dispatcher init
        :type oauth_token: :obj:`str`

        :return: True if result is ok
        '''
        skill_id, oauth_token = self._check_auth(skill_id, oauth_token)
        url = api.Methods.api_url(skill_id, api.Methods.IMAGES) + image_id
        result = await api.request(
            self.session, oauth_token,
            request_method='DELETE', custom_url=url
        )
        return result['result'] == 'ok' if 'result' in result else False



from attr import attrs, attrib
from . import AliceObject


@attrs
class EntityValue(AliceObject):
    """EntityValue object"""

    # YANDEX.FIO
    first_name = attrib(default=None, type=str)
    patronymic_name = attrib(default=None, type=str)
    last_name = attrib(default=None, type=str)

    # YANDEX.GEO
    country = attrib(default=None, type=str)
    city = attrib(default=None, type=str)
    street = attrib(default=None, type=str)
    house_number = attrib(default=None, type=str)
    airport = attrib(default=None, type=str)

    # YANDEX.DATETIME
    year = attrib(default=None, type=str)
    year_is_relative = attrib(default=False, type=bool)
    month = attrib(default=None, type=str)
    month_is_relative = attrib(default=False, type=bool)
    day = attrib(default=None, type=str)
    day_is_relative = attrib(default=False, type=bool)
    hour = attrib(default=None, type=str)
    hour_is_relative = attrib(default=False, type=bool)
    minute = attrib(default=None, type=str)
    minute_is_relative = attrib(default=False, type=bool)



from attr import attrs, attrib
from aioalice.utils import ensure_cls
from . import AliceObject, Card, Button


@attrs
class Response(AliceObject):
    """Response object"""

    text = attrib(type=str)
    tts = attrib(default=None, type=str)
    card = attrib(default=None, convert=ensure_cls(Card))
    buttons = attrib(default=None, convert=ensure_cls(Button))
    end_session = attrib(default=False, type=bool)



import asyncio
import logging
import functools
from aiohttp import web
from ..utils import json, generate_json_payload
from ..types import AliceRequest, AliceResponse, Response


log = logging.getLogger(__name__)


DEFAULT_WEB_PATH = '/alicewh/'
ALICE_DISPATCHER_KEY = 'ALICE_DISPATCHER'
ERROR_RESPONSE_KEY = 'ALICE_ERROR_RESPONSE'

DEFAULT_ERROR_RESPONSE_TEXT = 'Server error. Developer has to check logs.'
# Max time to response to API is 1.5s
# with server on Aruba (Italy) 1.2s is a critical timeout for whole processing
# NOTE that this timeout can help only if using non-blocking IO
# in e.g use asyncio.sleep instead of time.sleep, aiohttp instead of requests, etc
# Whole processing usually takes from 0.0004 до 0.001 (depends on system IO),
# but Yandex starts countdown as user asks a question, request processing takes some time
RESPONSE_TIMEOUT = 1.2


class WebhookRequestHandler(web.View):
    """
    Simple Wehhook request handler for aiohttp web server.

    You need to register that in app:

    .. code-block:: python3

        app.router.add_route('*', '/your/webhook/path', WebhookRequestHadler, name='webhook_handler')

    But first you need to configure application for getting Dispatcher instance from request handler!
    It must always be with key 'ALICE_DISPATCHER'

    .. code-block:: python3

        dp = Dispatcher()
        app['ALICE_DISPATCHER'] = dp

    """

    def get_dispatcher(self):
        """
        Get Dispatcher instance from environment
        """
        return self.request.app[ALICE_DISPATCHER_KEY]

    async def parse_request(self):
        """
        Read request from stream and deserialize it.
        :return: :class:`aioalice.types.AliceRequest`
        """
        data = await self.request.json()
        try:
            return AliceRequest(self.request, **data)
        except Exception:
            log.exception('Exception loading AliceRequest from\n%r', data)
            raise

    async def process_request(self, request):
        """
        You have to respond in less than 1.5 seconds to webhook.

        So... If you process longer than 1.2 (RESPONSE_TIMEOUT) seconds
        webhook automatically respond with FALLBACK VALUE (ERROR_RESPONSE_KEY)

        :param request:
        :return:
        """
        dispatcher = self.get_dispatcher()
        loop = dispatcher.loop

        # Analog of `asyncio.wait_for` but without cancelling task
        waiter = loop.create_future()
        timeout_handle = loop.call_later(RESPONSE_TIMEOUT, asyncio.tasks._release_waiter, waiter)
        done_cb = functools.partial(asyncio.tasks._release_waiter, waiter)

        fut = asyncio.ensure_future(dispatcher.process_request(request), loop=loop)
        fut.add_done_callback(done_cb)

        try:
            try:
                await waiter
            except asyncio.futures.CancelledError:
                fut.remove_done_callback(done_cb)
                fut.cancel()
                raise

            if fut.done():
                return fut.result()
            else:
                fut.remove_done_callback(done_cb)
                fut.add_done_callback(self.warn_slow_process(request))
        finally:
            timeout_handle.cancel()

    def warn_slow_process(self, request):
        """
        Wrapper for slow requests warning
        """

        def slow_request_processor(task):
            """
            Handle response after 1.2 sec (RESPONSE_TIMEOUT)

            :param task:
            :return:
            """
            log.warning('Long request processing detected.\n'
                        'Request was %s\nYou have to process request in %ss\n'
                        'request was automatically responded with `ERROR_RESPONSE_KEY`',
                        request, RESPONSE_TIMEOUT)

            dispatcher = self.get_dispatcher()
            loop = dispatcher.loop

            try:
                result = task.result()
            except Exception as e:
                log.info('Slow request processor raised an error, passing to errors_handlers')
                loop.create_task(dispatcher.errors_handlers.notify(dispatcher, request, e))
            else:
                log.warning('Result is %s', result)

        return slow_request_processor

    def default_error_response(self, alice_request):
        """
        Default error response will be called on timeout
        if processing of the request will take more than 1.2s (RESPONSE_TIMEOUT)

        :param result: dict or AliceRequest
        :return: AliceResponse
        """
        default_response = self.request.app[ERROR_RESPONSE_KEY]
        response = alice_request.response(default_response)
        return generate_json_payload(**response.to_json())

    def get_response(self, result, request):
        """
        Make response object from result.

        :param result: dict or AliceResponse
        :return:
        """
        if isinstance(result, AliceResponse):
            return generate_json_payload(**result.to_json())
        if result is None:
            log.critical('Got `None` instead of a response!\nGenerating'
                         ' default error response based on %r', request)
            return self.default_error_response(request)
        if not isinstance(result, dict):
            # If result is not a dict, it may cause an error. Warn developer
            log.warning('Result expected `AliceResponse` or dict, got %r (%r)',
                        type(result), result)
        return result

    async def post(self):
        """
        Process POST response

        :return: :class:`aiohttp.web.Response`
        """
        request = await self.parse_request()
        result = await self.process_request(request)
        # request has to be passed to generate fallback value
        # if None is returned from process_request or on timeout
        response = self.get_response(result, request)
        return web.json_response(response, dumps=json.dumps)


def configure_app(app, dispatcher, path=DEFAULT_WEB_PATH,
                  default_response_or_text=DEFAULT_ERROR_RESPONSE_TEXT):
    """
    You can prepare web.Application for working with webhook handler.

    :param app: :class:`aiohttp.web.Application`
    :param dispatcher: Dispatcher instance
    :param path: Path to your webhook.
    :default_response_or_text: `aioalice.types.Response` OR text to answer user on fail or timeout
    :return:
    """
    app.on_shutdown.append(dispatcher.shutdown)
    app.router.add_route('*', path, WebhookRequestHandler, name='alice_webhook_handler')
    app[ALICE_DISPATCHER_KEY] = dispatcher
    # Prepare default Response
    if isinstance(default_response_or_text, Response):
        app[ERROR_RESPONSE_KEY] = default_response_or_text
    elif isinstance(default_response_or_text, str):
        app[ERROR_RESPONSE_KEY] = Response(default_response_or_text)
    else:
        response_text = str(default_response_or_text)
        app[ERROR_RESPONSE_KEY] = Response(response_text)
        log.warning('Automatically converted default_response_or_text'
                    ' to str\nIt\'ll be %r\nConsider using string or'
                    ' `aioalice.types.Response` next time', response_text)


def get_new_configured_app(dispatcher, path=DEFAULT_WEB_PATH,
                           default_response_or_text=DEFAULT_ERROR_RESPONSE_TEXT):
    """
    Create new :class:`aiohttp.web.Application` and configure it.

    :param dispatcher: Dispatcher instance
    :param path: Path to your webhook.
    :default_response_or_text: `aioalice.types.Response` OR text to answer user on fail or timeout
    :return:
    """
    app = web.Application()
    configure_app(app, dispatcher, path, default_response_or_text)
    return app



from attr import attrs, attrib

from aioalice.utils import ensure_cls
from . import AliceObject, MediaButton


@attrs
class Image(AliceObject):
    """Image object"""
    image_id = attrib(type=str)
    title = attrib(type=str)
    description = attrib(type=str)
    button = attrib(default=None, convert=ensure_cls(MediaButton))



import asyncio

from .dispatcher import Dispatcher
from .dispatcher.webhook import get_new_configured_app, configure_app

try:
    import uvloop
except ImportError:
    pass
else:
    asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())


__version__ = '1.2.2'



from attr import attrs, attrib

from aioalice.utils import ensure_cls
from . import AliceObject, MediaButton


@attrs
class CardFooter(AliceObject):
    """This object represents a card's footer"""
    text = attrib(type=str)
    button = attrib(default=None, convert=ensure_cls(MediaButton))



# https://gist.github.com/surik00/a6c2804a2d18a2ab75630bb5d93693c8

import inspect
import functools


def safe_kwargs(func_or_class):
    spec = inspect.getfullargspec(func_or_class)
    all_args = spec.args

    @functools.wraps(func_or_class)
    def wrap(*args, **kwargs):
        accepted_kwargs = {k: v for k, v in kwargs.items() if k in all_args}
        return func_or_class(*args, **accepted_kwargs)

    return wrap



from attr import attrs, attrib
from aiohttp.web import Request as WebRequest
from aioalice.utils import ensure_cls
from . import AliceObject, Meta, Session, \
    Card, Request, Response, AliceResponse


@attrs
class AliceRequest(AliceObject):
    """AliceRequest is a request from Alice API"""
    original_request = attrib(type=WebRequest)
    meta = attrib(convert=ensure_cls(Meta))
    request = attrib(convert=ensure_cls(Request))
    session = attrib(convert=ensure_cls(Session))
    version = attrib(type=str)

    def _response(self, response):
        return AliceResponse(
            response=response,
            session=self.session.base,
            version=self.version
        )

    def response(self, responose_or_text, **kwargs):
        """
        Generate response

        :param responose_or_text: Response or Response's text:
            if responose_or_text is not an instance of Response,
            it is passed to the Response initialisator with kwargs.
            Otherwise it is used as a Response

        :param kwargs: tts, card, buttons, end_session for Response
            NOTE: if you want to pass card, concider using one of
              these methods: response_big_image, response_items_list
        :return: AliceResponse
        """
        if not isinstance(responose_or_text, Response):
            responose_or_text = Response(responose_or_text, **kwargs)
        return self._response(responose_or_text)

    def response_big_image(self, text, image_id, title, description, button=None, **kwargs):
        """
        Generate response with Big Image card

        :param text: Response's text
        :param image_id: Image's id for BigImage Card
        :param title: Image's title for BigImage Card
        :param description: Image's description for BigImage Card
        :param button: Image's button for BigImage Card
        :param kwargs: tts, buttons, end_session for Response
        :return: AliceResponse
        """
        return self._response(
            Response(
                text, **kwargs,
                card=Card.big_image(image_id, title, description, button),
            )
        )

    def response_items_list(self, text, header, items, footer=None, **kwargs):
        """
        Generate response with Items List card

        :param text: Response's text
        :param header: Card's header
        :param items: Card's items - list of `Image` objects
        :param footer: Card's footer
        :param kwargs: tts, buttons, end_session for Response
        :return: AliceResponse
        """
        return self._response(
            Response(
                text, **kwargs,
                card=Card.items_list(header, items, footer)
            )
        )



from attr import attrs, attrib

from aioalice.utils import safe_kwargs, ensure_cls
from . import AliceObject, Interfaces


@safe_kwargs
@attrs
class Meta(AliceObject):
    """Meta object"""
    locale = attrib(type=str)
    timezone = attrib(type=str)
    client_id = attrib(type=str)
    interfaces = attrib(default=None, convert=ensure_cls(Interfaces))
    flags = attrib(factory=list)



class DialogsError(Exception):
    '''Base class for alice exceptions'''


class DialogsAPIError(Exception):
    '''Base Exception for API related requests'''
    __subclasses = []
    match = None

    def __init_subclass__(cls, match=None, **kwargs):
        super(DialogsAPIError, cls).__init_subclass__(**kwargs)
        if match is not None:
            cls.match = match.lower()
            cls.__subclasses.append(cls)

    @classmethod
    def detect(cls, description):
        """Detect API Error (match by response text)"""
        description = description.strip()
        match = description.lower()
        for err in cls.__subclasses:
            if err is cls:
                pass
            if err.match in match:
                raise err(description)
        raise cls(description)


class AuthRequired(Exception):
    '''Passed is skill_id of oauth_token is not provided'''


class ApiChanged(DialogsError):
    '''Is thrown if there are some unpredicted changes in API'''


class NetworkError(DialogsAPIError):
    '''Is thrown when aiohttp client throws an error'''


# class ClientError(DialogsAPIError):
#     '''Is thrown when response code is 4xx'''


# class ServerError(DialogsAPIError):
#     '''Is thrown when response code is 5xx'''


class Forbidden(DialogsAPIError, match='Forbidden'):
    '''Is thrown when Authorization failed (403)'''


class ContentNotProvided(DialogsAPIError, match='URL or FILE is needed'):
    '''Is thrown when no image is provided within request'''


class InvalidImageID(DialogsAPIError, match='Invalid image ID'):
    '''Is thrown if a wrong image_id is provided within request'''



from attr import attrs, attrib
from . import AliceObject


@attrs
class Button(AliceObject):
    """Button object"""
    title = attrib(type=str)
    url = attrib(default=None, type=str)
    payload = attrib(default=None)
    hide = attrib(default=True, type=bool)



from attr import attrs, attrib

from aioalice.utils import safe_kwargs, ensure_cls
from aioalice.utils.helper import Helper, HelperMode, Item
from . import AliceObject, Markup, NaturalLanguageUnderstanding


@safe_kwargs
@attrs
class Request(AliceObject):
    """Request object"""
    type = attrib(type=str)
    command = attrib(default='', type=str)  # Can be none if payload passed
    original_utterance = attrib(default='', type=str)  # Can be none if payload passed
    markup = attrib(default=None)
    payload = attrib(default=None)
    nlu = attrib(default=None, convert=ensure_cls(NaturalLanguageUnderstanding))

    @type.validator
    def check(self, attribute, value):
        """
        Type can be 'SimpleUtterance' or 'ButtonPressed'
            "SimpleUtterance" — голосовой ввод;
            "ButtonPressed" — нажатие кнопки.
        """
        if value not in RequestType.all():
            raise ValueError(f'Request type must be "SimpleUtterance" or "ButtonPressed", not "{value}"')

    def __attrs_post_init__(self):
        if self.markup is not None:
            self.markup = Markup(**self.markup)


class RequestType(Helper):
    mode = HelperMode.CamelCase

    SIMPLE_UTTERANCE = Item()  # SimpleUtterance
    BUTTON_PRESSED = Item()  # ButtonPressed



from attr import attrs, attrib

from aioalice.utils import ensure_cls
from aioalice.utils.helper import Helper, HelperMode, Item
from . import AliceObject, MediaButton, Image, CardHeader, CardFooter


@attrs
class Card(AliceObject):
    """This object represents a Card either of type `BigImage` or `ItemsList`"""

    type = attrib(type=str)

    # for BigImage
    image_id = attrib(default=None, type=str)
    title = attrib(default=None, type=str)
    description = attrib(default=None, type=str)
    button = attrib(default=None, convert=ensure_cls(MediaButton))

    # for ItemsList
    header = attrib(default=None, convert=ensure_cls(CardHeader))
    items = attrib(default=None, convert=ensure_cls(Image))  # List of Image objects

    footer = attrib(default=None, convert=ensure_cls(CardFooter))

    @type.validator
    def check(self, attribute, value):
        """
        Type can be 'BigImage' or 'ItemsList'
            "BigImage" — с одним изображением
            "ItemsList" — с галереей из нескольких изображений
        """
        if value not in CardType.all():
            raise ValueError(f'Card type must be "BigImage" or "ItemsList", not "{value}"')

    @classmethod
    def big_image(cls, image_id, title, description, button=None):
        """
        Generate Big Image card

        :param image_id: Image's id for BigImage Card
        :param title: Image's title for BigImage Card
        :param description: Image's description for BigImage Card
        :param button: Image's button for BigImage Card
        :return: Card
        """
        return cls(
            CardType.BIG_IMAGE,
            image_id=image_id,
            title=title,
            description=description,
            button=button,
        )

    @classmethod
    def items_list(cls, header, items, footer=None):
        """
        Generate Items List card

        :param header: Card's header
        :param items: Card's items - list of `Image` objects
        :param footer: Card's footer
        :return: Card
        """
        return cls(
            CardType.ITEMS_LIST,
            header=header,
            items=items,
            footer=footer,
        )


class CardType(Helper):
    mode = HelperMode.CamelCase

    BIG_IMAGE = Item()  # BigImage
    ITEMS_LIST = Item()  # ItemsList



import re
import inspect
import logging

from ..utils.helper import Helper, HelperMode, Item

log = logging.getLogger(__name__)


async def check_filter(filter_, args):
    """
    Helper for executing filter

    :param filter_:
    :param args:
    :param kwargs:
    :return:
    """
    if not callable(filter_):
        raise TypeError(f'Filter must be callable and/or awaitable! Error with {filter_}')

    if inspect.isawaitable(filter_) or inspect.iscoroutinefunction(filter_):
        return await filter_(*args)
    else:
        return filter_(*args)


async def check_filters(filters, args):
    """
    Check list of filters

    :param filters:
    :param args:
    :return:
    """
    if filters is not None:
        for f in filters:
            filter_result = await check_filter(f, args)
            if not filter_result:
                return False
    return True


class Filter:
    """
    Base class for filters
    """

    def __call__(self, *args, **kwargs):
        return self.check(*args, **kwargs)

    def check(self, *args, **kwargs):
        raise NotImplementedError


class AsyncFilter(Filter):
    """
    Base class for asynchronous filters
    """

    def __aiter__(self):
        return None

    def __await__(self):
        return self.check

    async def check(self, *args, **kwargs):
        raise NotImplementedError


class StringCompareFilter(AsyncFilter):
    """
    Base for string comparison
    """

    def __init__(self, lines):
        self.lines = [w.lower() for w in lines]


class StartsWithFilter(StringCompareFilter):
    """
    Check if command starts with one of these lines
    """

    async def check(self, areq):
        command = areq.request.command.lower()
        return any([command.startswith(line) for line in self.lines])


class ContainsFilter(StringCompareFilter):
    """
    Check if command contains one of these lines
    """

    async def check(self, areq):
        command = areq.request.command.lower()
        return any([line in command for line in self.lines])


class CommandsFilter(AsyncFilter):
    """
    Check if command is one of these phrases
    Pass commands in lower case
    """

    def __init__(self, commands):
        self.commands = commands

    async def check(self, areq):
        command = areq.request.command.lower()
        return command in self.commands


class StateFilter(AsyncFilter):
    """Check user's state"""

    def __init__(self, dispatcher, state):
        self.dispatcher = dispatcher
        self.state = state

    async def check(self, areq):
        if self.state == '*':
            return True
        user_state = await self.dispatcher.storage.get_state(areq.session.user_id)
        return user_state == self.state


class StatesListFilter(StateFilter):
    """Check if user's state is in list of states"""

    async def check(self, areq):
        user_state = await self.dispatcher.storage.get_state(areq.session.user_id)
        return user_state in self.state


class RegexpFilter(Filter):
    """
    Regexp filter for original_utterance (full request text)
    If `AliceRequest.request.original_utterance` matches regular expression
    """

    def __init__(self, regexp):
        self.regexp = re.compile(regexp, flags=re.IGNORECASE | re.MULTILINE)

    def check(self, areq):
        return bool(self.regexp.search(areq.request.original_utterance))


class RequestTypeFilter(Filter):
    """
    Check AliceRequest.request type
    On API 1.0 it can be 'SimpleUtterance' or 'ButtonPressed'
    """

    def __init__(self, content_types):
        if isinstance(content_types, str):
            content_types = [content_types]
        self.content_types = content_types

    def check(self, areq):
        return areq.request.type in self.content_types


class ExceptionsFilter(Filter):
    """
    Filter for exceptions
    """

    def __init__(self, exception):
        self.exception = exception

    def check(self, dispatcher, update, exception):
        return isinstance(exception, self.exception)


def generate_default_filters(dispatcher, *args, **kwargs):
    """
    Prepare filters

    :param dispatcher: for states
    :param args:
    :param kwargs:
    :return:
    """
    filters_list = []

    for name, filter_data in kwargs.items():
        if filter_data is None:
            # skip not setted filter names
            # Note that state by default is not None,
            # check dispatcher.storage for more information
            continue

        if name == DefaultFilters.REQUEST_TYPE:
            filters_list.append(RequestTypeFilter(filter_data))
        elif name == DefaultFilters.COMMANDS:
            if isinstance(filter_data, str):
                filters_list.append(CommandsFilter([filter_data]))
            else:
                filters_list.append(CommandsFilter(filter_data))
        elif name == DefaultFilters.STARTS_WITH:
            if isinstance(filter_data, str):
                filters_list.append(StartsWithFilter([filter_data]))
            else:
                filters_list.append(StartsWithFilter(filter_data))
        elif name == DefaultFilters.CONTAINS:
            if isinstance(filter_data, str):
                filters_list.append(ContainsFilter([filter_data]))
            else:
                filters_list.append(ContainsFilter(filter_data))
        elif name == DefaultFilters.STATE:
            if isinstance(filter_data, (list, set, tuple, frozenset)):
                filters_list.append(StatesListFilter(dispatcher, filter_data))
            else:
                filters_list.append(StateFilter(dispatcher, filter_data))
        elif name == DefaultFilters.FUNC:
            filters_list.append(filter_data)
        elif name == DefaultFilters.REGEXP:
            filters_list.append(RegexpFilter(filter_data))
        elif isinstance(filter_data, Filter):
            filters_list.append(filter_data)
        else:
            log.warning('Unexpected filter with name %r of type `%r` (%s)',
                        name, type(filter_data), filter_data)

    filters_list += list(args)  # Some custom filters

    return filters_list


class DefaultFilters(Helper):
    mode = HelperMode.snake_case

    REQUEST_TYPE = Item()  # request_type
    STARTS_WITH = Item()  # starts_with
    CONTAINS = Item()  # contains
    COMMANDS = Item()  # commands
    REGEXP = Item()  # regexp
    STATE = Item()  # state
    FUNC = Item()  # func



from attr import attrs, attrib
from . import AliceObject


@attrs
class UploadedImage(AliceObject):
    """This object represents an uploaded image"""
    id = attrib(type=str)
    origUrl = attrib(default=None, type=str)
    # origUrl will be None if image was uploaded from bytes, not by url

    @property
    def orig_url(self):
        return self.origUrl



from . import exceptions
from .json import json
from .payload import generate_json_payload
from .safe_kwargs import safe_kwargs


def ensure_cls(klass):
    def converter(val):
        if val is None:
            return
        if isinstance(val, dict):
            return klass(**val)
        if isinstance(val, list):
            return [converter(v) for v in val]
        if not isinstance(val, klass):
            return klass(val)
        return val
    return converter



from .base import AliceObject
from .interfaces import Interfaces
from .meta import Meta
from .markup import Markup
from .entity_tokens import EntityTokens
from .entity_value import EntityValue
from .entity import Entity, EntityType
from .natural_language_understanding import NaturalLanguageUnderstanding
from .request import Request, RequestType
from .session import BaseSession, Session

from .quota import Quota
from .uploaded_image import UploadedImage
from .media_button import MediaButton
from .image import Image
from .card_header import CardHeader
from .card_footer import CardFooter
from .card import Card, CardType

from .button import Button
from .response import Response

from .alice_response import AliceResponse
from .alice_request import AliceRequest



DEFAULT_FILTER = ['self', 'cls']


def generate_json_payload(exclude=[], **kwargs):
    """
    Generate payload

    Usage: payload = generate_json_payload(**locals(), exclude=['foo'])

    :param exclude:
    :param kwargs:
    :return: dict
    """
    return {key: _normalize(value) for key, value in kwargs.items() if
            key not in exclude + DEFAULT_FILTER
            and value is not None
            and not key.startswith('_')}


def _normalize(obj):
    """
    Normalize dicts and lists

    :param obj:
    :return: normalized object
    """
    if isinstance(obj, dict):
        return {k: _normalize(v) for k, v in obj.items() if v is not None}
    # elif hasattr(obj, 'to_json'):
    #     return _normalize(obj.to_json())
    elif isinstance(obj, list):
        return [_normalize(item) for item in obj]
    return obj



# Natural Language Understanding: https://medium.com/@lola.com/nlp-vs-nlu-whats-the-difference-d91c06780992
from attr import attrs, attrib
from aioalice.utils import ensure_cls
from . import AliceObject, Entity


@attrs
class NaturalLanguageUnderstanding(AliceObject):
    """Natural Language Understanding object"""
    tokens = attrib(factory=list)
    entities = attrib(factory=list, convert=ensure_cls(Entity))



from attr import attrs, attrib
from aioalice.utils import ensure_cls
from . import AliceObject, BaseSession, Response


@attrs
class AliceResponse(AliceObject):
    """AliceResponse is a response to Alice API"""

    response = attrib(convert=ensure_cls(Response))
    session = attrib(convert=ensure_cls(BaseSession))
    version = attrib(type=str)



from attr import attrs, attrib
from . import AliceObject


@attrs
class Quota(AliceObject):
    """Quota object. Values in bytes"""
    total = attrib(type=int)
    used = attrib(type=int)
    available = attrib(init=False, default=0)

    def __attrs_post_init__(self):
        self.available = self.total - self.used



import logging

# This state will be default instead of `None` to gain maximum speed
# So if handler is registered with `state=None`,
# no State filter will be applied by default
# use `DEFAULT_STATE` for FSM, else no state check will be applied
DEFAULT_STATE = 'DEFAULT_STATE'

log = logging.getLogger(__name__)


class BaseStorage:
    """
    You are able to save current user's state
    and data for all steps in states storage
    """

    async def close(self):
        """
        You have to override this method and use when application shutdowns.
        Perhaps you would like to save data and etc.

        :return:
        """
        raise NotImplementedError

    async def wait_closed(self):
        """
        You have to override this method for all asynchronous storages (e.g., Redis).

        :return:
        """
        raise NotImplementedError

    async def get_state(self, user_id):
        """
        Get current state of user.
        Default is `DEFAULT_STATE`

        :param user_id:
        :param default:
        :return:
        """
        raise NotImplementedError

    async def get_data(self, user_id):
        """
        Get state data for user.
        Default is `{}`

        :param user_id:
        :param default:
        :return:
        """
        raise NotImplementedError

    async def set_state(self, user_id, state):
        """
        Set new state for user

        :param user_id:
        :param state:
        """
        raise NotImplementedError

    async def set_data(self, user_id, data):
        """
        Set data for user

        :param user_id:
        :param data:
        """
        raise NotImplementedError

    async def update_data(self, user_id, data=None, **kwargs):
        """
        Update data for user

        You can use data parameter or|and kwargs.

        :param data:
        :param user_id:
        :param kwargs:
        :return:
        """
        raise NotImplementedError

    async def reset_data(self, user_id):
        """
        Reset data for user.

        :param user_id:
        :return:
        """
        await self.set_data(user_id, data={})

    async def reset_state(self, user_id, with_data=False):
        """
        Reset state for user.
        You may desire to use this method when finishing conversations.

        :param user_id:
        :param with_data:
        :return:
        """
        await self.set_state(user_id, state=DEFAULT_STATE)
        if with_data:
            await self.reset_data(user_id)

    async def finish(self, user_id):
        """
        Finish conversation with user.

        :param user_id:
        :return:
        """
        await self.reset_state(user_id, with_data=True)


class DisabledStorage(BaseStorage):
    """
    Empty storage. Use it if you don't want to use Finite-State Machine
    """

    async def close(self):
        pass

    async def wait_closed(self):
        pass

    async def get_state(self, user_id):
        return DEFAULT_STATE

    async def get_data(self, user_id):
        self._warn()
        return {}

    async def set_state(self, user_id, state):
        self._warn()

    async def set_data(self, user_id, data):
        self._warn()

    async def update_data(self, user_id, data=None, **kwargs):
        self._warn()

    @staticmethod
    def _warn():
        log.warning("You haven’t set any storage yet so no states and no data will be saved.\n"
                    "You can connect MemoryStorage for debug purposes or non-essential data.")


class MemoryStorage(BaseStorage):
    """In-memory states storage"""

    def __init__(self):
        self.data = {}

    async def close(self):
        self.data.clear()

    async def wait_closed(self):
        pass

    def _get_user_data(self, user_id):
        if user_id not in self.data:
            self.data[user_id] = {'state': DEFAULT_STATE, 'data': {}}
        return self.data[user_id]

    async def get_state(self, user_id):
        user = self._get_user_data(user_id)
        return user['state']

    async def get_data(self, user_id):
        user = self._get_user_data(user_id)
        return user['data']

    async def set_state(self, user_id, state):
        user = self._get_user_data(user_id)
        user['state'] = state

    async def set_data(self, user_id, data):
        user = self._get_user_data(user_id)
        user['data'] = data

    async def update_data(self, user_id, data=None, **kwargs):
        user = self._get_user_data(user_id)
        if data is None:
            data = {}
        user['data'].update(data, **kwargs)



import logging
from attr import attrs, attrib

from aioalice.utils import ensure_cls
from aioalice.utils.helper import Helper, HelperMode, Item
from . import AliceObject, EntityTokens, EntityValue

log = logging.getLogger(__name__)


@attrs
class Entity(AliceObject):
    """Entity object"""
    type = attrib(type=str)
    tokens = attrib(convert=ensure_cls(EntityTokens))
    value = attrib(factory=dict)

    @type.validator
    def check(self, attribute, value):
        """Report unknown type"""
        if value not in EntityType.all():
            log.error('Unknown Entity type! `%r`', value)

    def __attrs_post_init__(self):
        """If entity type not number, convert to EntityValue"""
        if self.value and self.type != EntityType.YANDEX_NUMBER:
            self.value = EntityValue(**self.value)


class EntityType(Helper):
    mode = HelperMode.UPPER_DOT_SEPARATED

    YANDEX_GEO = Item()
    YANDEX_FIO = Item()
    YANDEX_NUMBER = Item()
    YANDEX_DATETIME = Item()



from attr import attrs, attrib
from . import AliceObject


@attrs
class MediaButton(AliceObject):
    """MediaButton object"""
    text = attrib(type=str)
    url = attrib(type=str)
    payload = attrib(default=None)



#!/usr/bin/env python3
import pathlib
import sys

from setuptools import find_packages, setup

try:
    from pip.req import parse_requirements
except ImportError:  # pip >= 10.0.0
    from pip._internal.req import parse_requirements

WORK_DIR = pathlib.Path(__file__).parent

# Check python version
MINIMAL_PY_VERSION = (3, 6)
if sys.version_info < MINIMAL_PY_VERSION:
    raise RuntimeError('aioAlice works only with Python {}+'.format('.'.join(map(str, MINIMAL_PY_VERSION))))

__version__ = '1.2.2'


def get_description():
    """
    Read full description from 'README-pypa.md'

    :return: description
    :rtype: str
    """
    with open('README-pypa.md', 'r', encoding='utf-8') as f:
        return f.read()


def get_requirements(filename=None):
    """
    Read requirements from 'requirements.txt'

    :return: requirements
    :rtype: list
    """
    if filename is None:
        filename = 'requirements.txt'

    file = WORK_DIR / filename

    install_reqs = parse_requirements(str(file), session='hack')
    return [str(ir.req) for ir in install_reqs]


setup(
    name='aioAlice',
    version=__version__,
    packages=find_packages(exclude=('tests', 'tests.*', 'examples',)),
    url='https://github.com/surik00/aioalice',
    license='MIT',
    author='Suren Khorenyan',
    requires_python='>=3.6',
    author_email='surenkhorenyan@gmail.com',
    description='Asynchronous library for Yandex Dialogs (Alice) API',
    long_description=get_description(),
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
    ],
    install_requires=get_requirements()
)



from attr import attrs, attrib
from . import AliceObject


@attrs
class BaseSession(AliceObject):
    """Base Session object"""
    session_id = attrib(type=str)
    message_id = attrib(type=int)
    user_id = attrib(type=str)


@attrs
class Session(BaseSession):
    """Session object"""
    new = attrib(type=bool)
    skill_id = attrib(type=str)

    @property
    def base(self):
        return BaseSession(
            self.session_id,
            self.message_id,
            self.user_id
        )



from .filters import check_filters


class SkipHandler(BaseException):
    """Raise this exception if the handler needs to be skipped"""


class Handler:

    def __init__(self):
        self.handlers = []

    def register(self, handler, filters=None, index=None):
        """
        Register callback

        Filters can be awaitable or not.

        :param handler: coroutine
        :param filters: list of filters
        :param index: you can reorder handlers
        """
        if filters and not isinstance(filters, (list, tuple, set)):
            filters = [filters]
        record = (filters, handler)
        if index is None:
            self.handlers.append(record)
        else:
            self.handlers.insert(index, record)

    def unregister(self, handler):
        """
        Remove handler

        :param handler: callback
        :return:
        """
        for handler_with_filters in self.handlers:
            _, registered = handler_with_filters
            if handler is registered:
                self.handlers.remove(handler_with_filters)
                return True
        raise ValueError('This handler is not registered!')

    async def notify(self, *args):
        """
        Notify handlers

        :param args:
        :return: instance of AliceResponse
            You *have* to return something to answer to API
            Consider returning AliceResponse or prepared JSON
        """

        for filters, handler in self.handlers:
            if await check_filters(filters, args):
                try:
                    return await handler(*args)
                except SkipHandler:
                    continue



# taken from aiogram: https://github.com/aiogram/aiogram/blob/master/aiogram/utils/helper.py

"""
Example:
    >>> from aioalice.utils.helper import Helper, ListItem, HelperMode, Item
    >>> class MyHelper(Helper):
    ...     mode = HelperMode.lowerCamelCase
    ...     FOO_ITEM = ListItem()
    ...     BAR_ITEM = ListItem()
    ...     BAZ_ITEM = ListItem()
    ...     LOREM = Item()
    ...
    >>> print(MyHelper.FOO_ITEM & MyHelper.BAR_ITEM)
    <<<  ['fooItem', 'barItem']
    >>> print(MyHelper.all())
    <<<  ['barItem', 'bazItem', 'fooItem', 'lorem']
"""


class Helper:
    mode = ''

    @classmethod
    def all(cls):
        """
        Get all consts
        :return: list
        """
        result = []
        for name in dir(cls):
            if not name.isupper():
                continue
            value = getattr(cls, name)
            if isinstance(value, ItemsList):
                result.append(value[0])
            else:
                result.append(value)
        return result


class HelperMode(Helper):
    mode = 'original'

    SCREAMING_SNAKE_CASE = 'SCREAMING_SNAKE_CASE'
    lowerCamelCase = 'lowerCamelCase'
    CamelCase = 'CamelCase'
    snake_case = 'snake_case'
    lowercase = 'lowercase'
    UPPER_DOT_SEPARATED = 'UPPER.DOT.SEPARATED'

    @classmethod
    def all(cls):
        return [
            cls.SCREAMING_SNAKE_CASE,
            cls.lowerCamelCase,
            cls.CamelCase,
            cls.snake_case,
            cls.lowercase,
        ]

    @classmethod
    def _screaming_snake_case(cls, text):
        """
        Transform text to SCREAMING_SNAKE_CASE

        :param text:
        :return:
        """
        if text.isupper():
            return text
        result = ''
        for pos, symbol in enumerate(text):
            if symbol.isupper() and pos > 0:
                result += '_' + symbol
            else:
                result += symbol.upper()
        return result

    @classmethod
    def _snake_case(cls, text):
        """
        Transform text to snake cale (Based on SCREAMING_SNAKE_CASE)

        :param text:
        :return:
        """
        if text.islower():
            return text
        return cls._screaming_snake_case(text).lower()

    @classmethod
    def _camel_case(cls, text, first_upper=False):
        """
        Transform text to camelCase or CamelCase

        :param text:
        :param first_upper: first symbol must be upper?
        :return:
        """
        result = ''
        need_upper = False
        for pos, symbol in enumerate(text):
            if symbol == '_' and pos > 0:
                need_upper = True
            else:
                if need_upper:
                    result += symbol.upper()
                else:
                    result += symbol.lower()
                need_upper = False
        if first_upper:
            result = result[0].upper() + result[1:]
        return result

    @classmethod
    def apply(cls, text, mode):
        """
        Apply mode for text

        :param text:
        :param mode:
        :return:
        """
        if mode == cls.UPPER_DOT_SEPARATED:
            return cls._screaming_snake_case(text).replace('_', '.')
        elif mode == cls.SCREAMING_SNAKE_CASE:
            return cls._screaming_snake_case(text)
        elif mode == cls.snake_case:
            return cls._snake_case(text)
        elif mode == cls.lowerCamelCase:
            return cls._camel_case(text)
        elif mode == cls.CamelCase:
            return cls._camel_case(text, True)
        elif mode == cls.lowercase:
            return cls._snake_case(text).replace('_', '')
        elif callable(mode):
            return mode(text)
        return text


class Item:
    """
    Helper item

    If a value is not provided,
    it will be automatically generated based on a variable's name
    """

    def __init__(self, value=None):
        self._value = value

    def __get__(self, instance, owner):
        return self._value

    def __set_name__(self, owner, name):
        if not name.isupper():
            raise NameError('Name for helper item must be in uppercase!')
        if not self._value:
            if hasattr(owner, 'mode'):
                self._value = HelperMode.apply(name, getattr(owner, 'mode'))


class ListItem(Item):
    """
    This item is always a list

    You can use &, | and + operators for that.
    """

    def add(self, other):
        return self + other

    def __get__(self, instance, owner):
        return ItemsList(self._value)

    def __getitem__(self, item):
        # Only for IDE. This method is never be called.
        return self._value

    # Need only for IDE
    __iadd__ = __add__ = __rand__ = __and__ = __ror__ = __or__ = add


class ItemsList(list):
    """
    Patch for default list

    This class provides +, &, |, +=, &=, |= operators for extending the list
    """

    def __init__(self, *seq):
        super(ItemsList, self).__init__(map(str, seq))

    def add(self, other):
        self.extend(other)
        return self

    __iadd__ = __add__ = __rand__ = __and__ = __ror__ = __or__ = add
